diff -ru dnsdist-1.9.3.orig/dnsdist-nghttp2-in.cc dnsdist-1.9.3/dnsdist-nghttp2-in.cc
--- dnsdist-1.9.3.orig/dnsdist-nghttp2-in.cc	2024-04-05 12:52:04.000000000 +0200
+++ dnsdist-1.9.3/dnsdist-nghttp2-in.cc	2024-04-23 15:55:23.146436261 +0200
@@ -222,8 +222,9 @@
 
 std::unique_ptr<DOHUnitInterface> IncomingHTTP2Connection::getDOHUnit(uint32_t streamID)
 {
-  // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-array-to-pointer-decay): clang-tidy is getting confused by assert()
-  assert(streamID <= std::numeric_limits<IncomingHTTP2Connection::StreamID>::max());
+  if (streamID > std::numeric_limits<IncomingHTTP2Connection::StreamID>::max()) {
+    throw std::runtime_error("Invalid stream ID while retrieving DoH unit");
+  }
   // NOLINTNEXTLINE(*-narrowing-conversions): generic interface between DNS and DoH with different types
   auto query = std::move(d_currentStreams.at(static_cast<IncomingHTTP2Connection::StreamID>(streamID)));
   return std::make_unique<IncomingDoHCrossProtocolContext>(std::move(query), std::dynamic_pointer_cast<IncomingHTTP2Connection>(shared_from_this()), streamID);
@@ -530,8 +531,9 @@
 
 IOState IncomingHTTP2Connection::sendResponse(const struct timeval& now, TCPResponse&& response)
 {
-  // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-array-to-pointer-decay): clang-tidy is getting confused by assert()
-  assert(response.d_idstate.d_streamID != -1);
+  if (response.d_idstate.d_streamID == -1) {
+    throw std::runtime_error("Invalid DoH stream ID while sending response");
+  }
   auto& context = d_currentStreams.at(response.d_idstate.d_streamID);
 
   uint32_t statusCode = 200U;
@@ -562,8 +564,10 @@
     return;
   }
 
-  // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-array-to-pointer-decay): clang-tidy is getting confused by assert()
-  assert(response.d_idstate.d_streamID != -1);
+  if (response.d_idstate.d_streamID == -1) {
+    throw std::runtime_error("Invalid DoH stream ID while handling I/O error notification");
+  }
+
   auto& context = d_currentStreams.at(response.d_idstate.d_streamID);
   context.d_buffer = std::move(response.d_buffer);
   sendResponse(response.d_idstate.d_streamID, context, 502, d_ci.cs->dohFrontend->d_customResponseHeaders);
diff -ru dnsdist-1.9.3.orig/dnsdist-tcp-downstream.cc dnsdist-1.9.3/dnsdist-tcp-downstream.cc
--- dnsdist-1.9.3.orig/dnsdist-tcp-downstream.cc	2024-04-05 12:52:04.000000000 +0200
+++ dnsdist-1.9.3/dnsdist-tcp-downstream.cc	2024-04-23 15:55:23.146436261 +0200
@@ -678,6 +678,7 @@
     /* we don't move the whole IDS because we will need for the responses to come */
     response.d_idstate.qtype = it->second.d_query.d_idstate.qtype;
     response.d_idstate.qname = it->second.d_query.d_idstate.qname;
+    response.d_idstate.d_streamID = it->second.d_query.d_idstate.d_streamID;
     DEBUGLOG("passing XFRresponse to client connection for "<<response.d_idstate.qname);
 
     it->second.d_query.d_xfrStarted = true;
diff -ru dnsdist-1.9.3.orig/dnsdist.cc dnsdist-1.9.3/dnsdist.cc
--- dnsdist-1.9.3.orig/dnsdist.cc	2024-04-05 12:52:04.000000000 +0200
+++ dnsdist-1.9.3/dnsdist.cc	2024-04-23 15:54:22.288561930 +0200
@@ -1670,9 +1670,18 @@
     /* we need an accurate ("real") value for the response and
        to store into the IDS, but not for insertion into the
        rings for example */
-    struct timespec now;
+    timespec now{};
     gettime(&now);
 
+    if ((dq.ids.qtype == QType::AXFR || dq.ids.qtype == QType::IXFR) && (dq.getProtocol() == dnsdist::Protocol::DoH || dq.getProtocol() == dnsdist::Protocol::DoQ || dq.getProtocol() == dnsdist::Protocol::DoH3)) {
+      dq.editHeader([](dnsheader& header) {
+        header.rcode = RCode::NotImp;
+        header.qr = true;
+        return true;
+      });
+      return processQueryAfterRules(dq, holders, selectedBackend);
+    }
+
     if (!applyRulesToQuery(holders, dq, now)) {
       return ProcessQueryResult::Drop;
     }
